<?php

use DataBase\IndexCollectionInterface;
use DataBase\IndexInterface;
use DataBase\MysqlIndexCollection;

/**
 * Class SiteSQL
 */
class SiteSQL
{

    const SQL_DEBUG = false;

    const SQL_NONE = 1;
    const SQL_ALL = 2;
    const SQL_INIT = 3;

    const SQL_ASSOC = 1;
    const SQL_INDEX = 2;

    /**
     * @var array
     */
    protected $structures = array();

    /**
     * @var IndexCollectionInterface[]
     */
    protected $indexes = [];

    /**
     * @var SiteSQL
     */
    static private $instance = null;
    /**
     * @var \mysqli
     */
    public $db = null;
    public $result = null;
    public $record = null;
    public $total = null;

    /**
     * Constructor
     */
    public function __construct(\mysqli $mysqli)
    {
        if ($mysqli->connect_error) {
            throw new InvalidArgumentException(sprintf("Erreur de connexion (%s), %s", $mysqli->connect_errno,$mysqli->connect_error));
        }

        $this->db = $mysqli;
        $this->db->set_charset("utf8");
    }

    /**
     * Get instance of the singleton
     *
     * @return SiteSQL
     */
    public static function getInstance()
    {
        if (!(self::$instance instanceof self)) {
            self::$instance = new self(new \mysqli(
                getenv('DB_HOST'),
                getenv('DB_USERNAME'),
                getenv('DB_PASSWORD'),
                getenv('DB_DATABASE')
            ));
        }
        return self::$instance;
    }

    /**
     * load the table structures from the cache
     *
     * @return void
     */
    public function loadStructures()
    {
        $phpCache = new \Cache\AppCache();

        $structCache = $phpCache->get('structures_' . $this->getDatabaseName());
        $this->structures = $structCache !== false ? $structCache : [];
    }

    /**
     * get the $structures property
     *
     * @return array
     */
    public function getStructures()
    {
        return $this->structures;
    }

    /**
     * close the connection
     *
     * @return bool
     */
    public function disconnect()
    {
        return $this->db->close();
    }

    /**
     * execute a statement SQL
     *
     * @param string $sql    the statement SQL
     * @param int    $type   type of result
     * @param int    $format format of array to return
     *
     * @return bool
     */
    public function query($sql, $type = self::SQL_NONE, $format = self::SQL_INDEX)
    {

        $dbg = Application::getDebugBar();
        if (getenv('APP_DEBUG') && !empty($dbg)) {
            $sqlLog = substr($sql, 0, 150);
            if ($sqlLog != $sql) {
                $sqlLog.= '...';
            }

            $dbg['time']->startMeasure('statement', $sqlLog);
        }

        $this->db->query("SET SQL_BIG_SELECTS=1");

        $this->record = array();

        $this->result = $this->db->query($sql);

        if (getenv('APP_DEBUG') && !empty($dbg)) {
            $dbg['time']->stopMeasure('statement');
        }

        if ($this->db->error) {
            if(getenv('APP_DEBUG') != 0){
                throw new DediException('SQL ERROR ' . $this->db->errno . ' : ' . $this->db->error);
            }
            error_log('SQL ERROR ' . $this->db->errno . ' : ' . $this->db->error);
            error_log((new Exception())->getTraceAsString());
            return false;
        }

        if ($this->result !== false && $this->result instanceof \mysqli_result) {
            $resulttype = ($format == self::SQL_INDEX) ? MYSQLI_NUM : MYSQLI_ASSOC;
            if ($type == self::SQL_ALL) {
                while ($row = $this->result->fetch_array($resulttype)) {
                    $this->record[] = $row;
                }
            } else {
                $this->record = $this->result->fetch_array($resulttype);
            }
            $this->result->free();
        }
        return true;
    }

    /**
     * Execute a SQL request and return all result
     *
     * @param string $sql the SQL request to execute
     *
     * @return array
     */
    public function getAll($sql)
    {
        $this->query($sql, self::SQL_ALL, self::SQL_ASSOC);
        return $this->record;
    }

    /**
     * Execute a SQL request and return the first result
     *
     * @param string $sql the SQL request to execute
     *
     * @return array
     */
    public function getOne($sql)
    {
        $this->query($sql, self::SQL_INIT, self::SQL_ASSOC);
        return $this->record;
    }

    /**
     * return the primary key of the table given
     *
     * @param string $table name of the table
     *
     * @return bool
     */
    public function getPrimaryKey($table)
    {
        $result = $this->getStructure($table);
        foreach ($result as $r) {
            if ($r['Key'] == 'PRI') {
                return $r['Field'];
            }
        }

        return false;
    }

    /**
     * return the struct of the table given
     *
     * @param string $table name of the table
     *
     * @return array
     */
    public function getStructure($table)
    {
        if ($this->structures === null) {
            $this->loadStructures();
        }

        if (empty($this->structures[$table])) {
            $this->structures[$table] = $this->getAll("DESC " . $table);
        }

        return $this->structures[$table];
    }

    /**
     * Fonction pour sauvegarder les données du tableau $data dans la table $table
     *
     * @param string $table nom de la table dans laquelle il faut insérer des data
     * @param array  $data  tableau de donées a insérer
     *
     * @return mixed
     * @throws DediException
     */
    public function save($table, $data)
    {

        $pk = $this->getPrimaryKey($table);

        if (empty($pk)) {
            throw new DediException('no primary key in table '. $table);
        }

        $autofill = array(
            $table . '_date_ajout' => 'NOW()',
            $table . '_date_modif' => 'NOW()',
        );

        $struct = $this->getStructure($table);

        $q = new Query();
        $action = (!empty($data[$pk])) ? 'update' : 'insert';
        $q->$action($table);

        foreach ($struct as $field) {
            if (in_array($field['Field'], array_keys($autofill))) {
                if ($field['Field'] == $table . '_date_ajout') {
                    if ($action == 'insert') {
                        $q->value($field['Field'], $autofill[$field['Field']]);
                    }
                } else {
                    $q->value($field['Field'], $autofill[$field['Field']]);
                }
            }

            if (!isset($data[$field['Field']])) {
                continue;
            }

            $q->value($field['Field'], $data[$field['Field']]);
        }

        if ($action == 'update') {
            $q->where($pk, $data[$pk]);
        }

        $this->query($q->build());

        if ($action == 'insert') {
            $data[$pk] = $this->db->insert_id;
        }

        return $data[$pk];
    }

    /**
     * Démarre une nouvelle transaction
     *
     * @return void
     */
    public function beginTransaction()
    {
        $this->db->autocommit(false);
    }

    /**
     * Fin de la transaction
     *
     * @return bool
     */
    public function endTransaction()
    {
        $result = $this->db->commit();
        $this->db->autocommit(true);

        return $result;
    }

    /**
     * @param string $table
     * @param array $data contenant les colonnes avec clés unique
     * @return array $data complété avec les données existantes si elles existent
     */
    public function merge($table, array $data)
    {
        $matches = $this->getUniqueIndexes($table)->getFiltered(
            function (IndexInterface $index) use ($data) {
                return $index->matchesData($data);
            }
        );

        foreach ($matches as $match) {
            $matchingData = $match->getMatchingData($data);
            $query = new Query();
            $query->select()->from($table);
            foreach ($matchingData as $column => $value) {
                $query->where($column, $value);
            }

            $foundData = $this->getOne($query->build());

            if ($foundData !== null) {
                return $data + $foundData;
            }
        }

        return $data;
    }

    /**
     * @param $table
     *
     * @return IndexCollectionInterface
     */
    private function getIndexes($table)
    {
        if (!array_key_exists($table, $this->indexes)) {
            $this->indexes[$table] = new MysqlIndexCollection($this->getAll("SHOW INDEX FROM `$table`"));
        }

        return $this->indexes[$table];
    }

    /**
     * @param $table
     *
     * @return IndexCollectionInterface
     */
    private function getUniqueIndexes($table)
    {
        return $this->getIndexes($table)->getUnique();
    }

    /**
     * @param string $table
     * @param array  $data
     *
     * @return mixed
     * @throws DediException
     */
    public function mergeAndSave($table, array $data)
    {
        return $this->save(
            $table,
            $this->merge(
                $table,
                $data
            )
        );
    }

    /**
     * @return mixed
     */
    public function getLastId()
    {
        return $this->db->insert_id;
    }

    /**
     * @return string
     */
    private function getDatabaseName()
    {
        return $this->getOne('SELECT DATABASE() as db')['db'];
    }
}
